/** \file Main.cpp
*/

/* Copyright (C) 2008-2009 Tomasz Ostrowski <tomasz.o.ostrowski at gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.

 * Miniscope is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <vcl.h>
#pragma hdrstop

#include "Main.h"
#include "display\DisplayOpenGL.h"
#include "display\ValueStringFormatter.h"
#include "FFT.h"
#include "Log.h"
#include "LogUnit.h"
#include "ScopedLock.h"
#include "Mutex.h"
#include "AboutUnit.h"

#include "device\MiniscopeDeviceInterface.h"

#include "SettingsUnit.h"
#include "Settings.h"

#include <SysUtils.hpp>
#include <math.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TfrmMain *frmMain;
Settings appSettings;

//---------------------------------------------------------------------------
__fastcall TfrmMain::TfrmMain(TComponent* Owner)
	: TForm(Owner), fZoomChangeGrid(1.25f), device(NULL)
{
	CLog::Instance()->SetFile(ChangeFileExt(Application->ExeName, ".log").c_str());
	CLog::Instance()->SetLevel(E_LOG_TRACE);

	display = new DisplayOpenGL(pnlCenter->Handle);
	display->SetMode(MODE_OSCILLOSCOPE);
	// some random settings
	const int iRange = 128;
	const float dy = 3.34e-4f;
	display->SetDiv(0.01f, dy);	// FS = 100Hz
	display->SetRange(iRange);	// 8bit -> +-128
    display->data = &data;
	btnModeScope->Down = true;

#ifdef CREATE_TEST_DATA
	// create and set some data
	data.clear();
	display->data = &data;
	for (int i=0; i<1024; i++)
	{
		data.push_back(	(float)iRange/2 * (	sin(2.0f * M_PI * (float)i/10.24f)           // ~0.1 FS (no leak),
											+ sin(2.0f * M_PI * (float)i/40.0f)         // 0.025 FS
											+ sin(2.0f * M_PI * (float)i/5.0f))+0.5f);	// 0.2 FS
	}
	// FFT
	float *ReIn = new float[data.size()];
	float *ReOut = new float[data.size()];
	float *ImOut = new float[data.size()];
	for(unsigned int i=0;i < data.size();i++)
		ReIn[i]= (float)data[i];
	FFT::Transform(data.size(), ReIn, NULL, ReOut, ImOut);
	dataFFT.resize(data.size()/2);
	//LOG(E_LOG_TRACE, "started");
	for(unsigned int i=0;i < data.size()/2;i++)
	{
		dataFFT[i] = sqrt((ReOut[i]*ReOut[i])+(ImOut[i]*ImOut[i]));
		//LOG(E_LOG_TRACE, "%f", dataFFT[i]);
	}
	delete [] ReIn;
	delete [] ReOut;
	delete [] ImOut;
#endif

	display->ChangeSize(pnlCenter->Width, pnlCenter->Height);

	// load custom cursors
	Screen->Cursors[CURSOR_HAND] = LoadCursor(HInstance, "HANDCUR");
	Screen->Cursors[CURSOR_ZOOM] = LoadCursor(HInstance, "ZOOMCUR");

	mouse_tool = MOUSE_CURSOR;
	pnlCenter->Cursor = crCross;
	btnMouseCursor->Down = true;

	bMovingPlot = false;

	displayOsc.fZoomX = 1.0;
	displayOsc.fZoomY = 1.0;
	displaySpe.fZoomX = 1.0;
	displaySpe.fZoomY = 1.0;

	float fValX, fValY;
	display->SetCursor1(0, pnlCenter->Height/2);
	display->PosToPhysical(0, pnlCenter->Height/2, fValX, fValY);
	UpdateCursorReadings(MODE_OSCILLOSCOPE, 1, fValX, fValY);
	display->SetCursor2(0, pnlCenter->Height/2);
	display->PosToPhysical(0, pnlCenter->Height/2, fValX, fValY);
	UpdateCursorReadings(MODE_OSCILLOSCOPE, 2, fValX, fValY);

	Timer->Interval = TrackBarSpeed->Position;	
}
//---------------------------------------------------------------------------

__fastcall TfrmMain::~TfrmMain()
{
	CLog::Instance()->Destroy();
}

void __fastcall TfrmMain::FormDestroy(TObject *Sender)
{
	delete display;
	if (device)
	{
		device->Disconnect();
		delete device;
		device = NULL;
	}

	AnsiString asConfigFile = ChangeFileExt( Application->ExeName, ".ini" );
	appSettings.iWidth = this->Width;
	appSettings.iHeight = this->Height;
	appSettings.iPosY = this->Top;
	appSettings.iPosX = this->Left;
	//appSettings.bSendOnEnter = btnSendOnEnter->Down;
	appSettings.Write(asConfigFile);
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::pnlCenterResize(TObject *Sender)
{
	if (display)
		display->ChangeSize(pnlCenter->Width, pnlCenter->Height);
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::TimerTimer(TObject *Sender)
{
	// timer seems to be better place for any error prone initialization
	// than constructor or OnShow event
	static bool once = false;
	if (!once)
	{
		once = true;
		// frmLog is already created - set callback to preview log entries
		CLog::Instance()->callbackLog = frmLog->OnLog;
		frmLog->callbackClose = CallbackCloseLogFn;
		// load device dlls
		AnsiString asDevDllDir = ExtractFileDir(Application->ExeName) + "\\device";
		DeviceInterface::EnumerateDlls(asDevDllDir);

		if (appSettings.asInputDeviceDllName != "")
		{
			AnsiString dllname = asDevDllDir + "\\" + appSettings.asInputDeviceDllName;
			for (unsigned int i=0; i<DeviceInterface::dlls.size(); i++)
			{
				if (DeviceInterface::dlls[i].name == dllname)
				{
					if (device)
						delete device;
					device = new DeviceInterface(dllname);
					if (device->Load())
					{
                    	ShowMessage("Failed loading device dll.");
					}
					device->callbackData = CallbackDataFn;

					display->SetDiv(device->capabilities->fSamplingPeriod[0],
									device->capabilities->fSensitivityLevels[0]);
					display->SetRange(pow(2, device->capabilities->iBitsPerSample - 1));
					display->ChangeSize(pnlCenter->Width, pnlCenter->Height);
					
					if (device->Connect())
					{
                    	ShowMessage("Failed connecting to device.");
					} else
					{
                    	device->AcquireStart();
					}
					break;
				}
			}
		}
	}

	ScopedLock<Mutex> lock(mutexDeviceData);
	if (bDataUpdate)
	{
		bDataUpdate = false;
		if (display->GetMode() == MODE_SPECTRUM)
			CalculateSpectrum();
	}
	display->Update(UPDATE_ALL);
}
//---------------------------------------------------------------------------

void TfrmMain::CalculateSpectrum(void)
{
	// FFT
	float *ReIn = new float[data.size()];
	float *ReOut = new float[data.size()];
	float *ImOut = new float[data.size()];
	for(unsigned int i=0;i < data.size();i++)
		ReIn[i]= (float)data[i];
	FFT::Transform(data.size(), ReIn, NULL, ReOut, ImOut);
	dataFFT.resize(data.size()/2);
	//LOG(E_LOG_TRACE, "started");
	for(unsigned int i=0;i < data.size()/2;i++)
	{
		dataFFT[i] = sqrt((ReOut[i]*ReOut[i])+(ImOut[i]*ImOut[i]));
		//LOG(E_LOG_TRACE, "%f", dataFFT[i]);
	}
	delete [] ReIn;
	delete [] ReOut;
	delete [] ImOut;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnStartStopClick(TObject *Sender)
{
	Timer->Enabled = !Timer->Enabled;	
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::TrackBarSpeedChange(TObject *Sender)
{
	Timer->Interval = TrackBarSpeed->Position;	
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::pnlCenterMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
	if (mouse_tool == MOUSE_CURSOR )
	{
		float fValX, fValY;
		if(Button==mbLeft)  //left button clicked
		{
			display->SetCursor1(X, Y);
			display->PosToPhysical(X, Y, fValX, fValY);
			UpdateCursorReadings(display->GetMode(), 1, fValX, fValY);
		}
		else if(Button==mbRight)
		{
			display->SetCursor2(X, Y);
			display->PosToPhysical(X, Y, fValX, fValY);
			UpdateCursorReadings(display->GetMode(), 2, fValX, fValY);
		}
	}
	else if (mouse_tool == MOUSE_OFFSET)
	{
		iPlotMoveStartX = X;
		iPlotMoveStartY = Y;
		bMovingPlot = true;
	}
	else if (mouse_tool == MOUSE_ZOOM || mouse_tool == MOUSE_ZOOM_X || mouse_tool == MOUSE_ZOOM_Y)
	{
		if(Button==mbLeft)  //left button clicked
		{
			if (mouse_tool != MOUSE_ZOOM_Y)
				DisplayState().fZoomX *= fZoomChangeGrid;
			if (mouse_tool != MOUSE_ZOOM_X)
				DisplayState().fZoomY *= fZoomChangeGrid;
		}
		else if(Button==mbRight)
		{
			if (mouse_tool != MOUSE_ZOOM_Y)
				DisplayState().fZoomX /= fZoomChangeGrid;
			if (mouse_tool != MOUSE_ZOOM_X)
				DisplayState().fZoomY /= fZoomChangeGrid;
		}
		display->SetZoom(X, Y, DisplayState().fZoomX, DisplayState().fZoomY);
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::pnlCenterMouseUp(TObject *Sender, TMouseButton Button,
	  TShiftState Shift, int X, int Y)
{
	if (bMovingPlot)
	{
		display->Move(iPlotMoveStartX, iPlotMoveStartY, X, Y);
		bMovingPlot = false;
	}
}
//---------------------------------------------------------------------------

void TfrmMain::UpdateCursorReadings(enum E_DISPLAY_MODE mode, int cursorId, float fValX, float fValY)
{
	static float fValX1 = 0.0, fValY1 = 0.0, fValX2 = 0.0, fValY2 = 0.0;

	if (mode==MODE_OSCILLOSCOPE)
	{
		if (cursorId==1)
		{
			StatusBar->Panels->Items[0]->Text =
				(ValueStringFormatter::TimeToStr(fValX, 3) + "  " +
				ValueStringFormatter::VoltageToStr(fValY, 3)).c_str();
			fValX1 = fValX;
			fValY1 = fValY;
		}
		else if (cursorId==2)
		{
			StatusBar->Panels->Items[1]->Text =
				(ValueStringFormatter::TimeToStr(fValX, 3) + "  " +
				ValueStringFormatter::VoltageToStr(fValY, 3)).c_str();
			fValX2 = fValX;
			fValY2 = fValY;
		}
		float dx = fabs(fValX1-fValX2);
		float dy = fabs(fValY1-fValY2);
		StatusBar->Panels->Items[2]->Text =
			((std::string)"dx=" + ValueStringFormatter::TimeToStr(dx, 3) + "  dy=" +
			ValueStringFormatter::VoltageToStr(dy, 3)).c_str();
		if (dx != 0)
		{
			StatusBar->Panels->Items[2]->Text = StatusBar->Panels->Items[2]->Text +
				"  1/dx=" + ValueStringFormatter::FreqToStr(1/dx, 3).c_str();
		}
	}
	else
	{
		if (cursorId==1)
		{
			StatusBar->Panels->Items[0]->Text =
				(ValueStringFormatter::FreqToStr(fValX, 3) + "  " +
				ValueStringFormatter::VoltageToStr(fValY, 3)).c_str();
			fValX1 = fValX;
			fValY1 = fValY;
		}
		else if (cursorId==2)
		{
			StatusBar->Panels->Items[1]->Text =
				(ValueStringFormatter::FreqToStr(fValX, 3) + "  " +
				ValueStringFormatter::VoltageToStr(fValY, 3)).c_str();
			fValX2 = fValX;
			fValY2 = fValY;
		}
		float dx = fabs(fValX1-fValX2);
		float dy = fabs(fValY1-fValY2);
		StatusBar->Panels->Items[2]->Text =
			((std::string)"dx=" + ValueStringFormatter::FreqToStr(dx, 3) + "  dy=" +
			ValueStringFormatter::VoltageToStr(dy, 3)).c_str();
	}
}


void __fastcall TfrmMain::ActionChangeMouseModeCursorExecute(TObject *Sender)
{
	mouse_tool = MOUSE_CURSOR;
	pnlCenter->Cursor = crCross;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionChangeMouseModeOffsetExecute(TObject *Sender)
{
	mouse_tool = MOUSE_OFFSET;
	pnlCenter->Cursor = (TCursor)CURSOR_HAND;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionChangeMouseModeZoomExecute(TObject *Sender)
{
	mouse_tool = MOUSE_ZOOM;
	pnlCenter->Cursor = (TCursor)CURSOR_ZOOM;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionChangeMouseModeZoomXExecute(TObject *Sender)
{
	mouse_tool = MOUSE_ZOOM_X;
	pnlCenter->Cursor = (TCursor)CURSOR_ZOOM;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionChangeMouseModeZoomYExecute(TObject *Sender)
{
	mouse_tool = MOUSE_ZOOM_Y;	
	pnlCenter->Cursor = (TCursor)CURSOR_ZOOM;	
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionZoomClearExecute(TObject *Sender)
{
	if (display->GetMode() == MODE_OSCILLOSCOPE)
	{
		displayOsc.fZoomX = 1.0;
		displayOsc.fZoomY = 1.0;
		display->SetZoom(-1, -1, displayOsc.fZoomX, displayOsc.fZoomY);
	}
	else
	{
		displaySpe.fZoomX = 1.0;
		displaySpe.fZoomY = 1.0;
		display->SetZoom(-1, -1, displaySpe.fZoomX, displaySpe.fZoomY);
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionModeScopeExecute(TObject *Sender)
{
	bDataUpdate = true;
	display->data = &data;
	display->SetMode(MODE_OSCILLOSCOPE);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionModeSpectrumExecute(TObject *Sender)
{
	bDataUpdate = true;
	display->data = &dataFFT;
	display->SetMode(MODE_SPECTRUM);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miAboutClick(TObject *Sender)
{
	AboutBox->ShowModal();
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::miCommonSettingsClick(TObject *Sender)
{
	frmSettings->appSettings = &appSettings;
	frmSettings->ShowModal();
	if (appSettings.bAlwaysOnTop)
		this->FormStyle = fsStayOnTop;
	else
		this->FormStyle = fsNormal;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::FormCreate(TObject *Sender)
{
    AnsiString asConfigFile = ChangeFileExt( Application->ExeName, ".ini" );
	appSettings.Read(asConfigFile);
	this->Width = appSettings.iWidth;
	this->Height = appSettings.iHeight;
	this->Top = appSettings.iPosY;
	this->Left = appSettings.iPosX;
	if (appSettings.bAlwaysOnTop)
		this->FormStyle = fsStayOnTop;
	else
		this->FormStyle = fsNormal;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miDeviceSettingsClick(TObject *Sender)
{
	if (device)
		device->ShowSettings();	
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miViewLogClick(TObject *Sender)
{
	frmLog->Visible = miViewLog->Checked;
	if (frmLog->Visible)
		frmLog->BringToFront();	
}
//---------------------------------------------------------------------------

void TfrmMain::CallbackCloseLogFn(void)
{
	miViewLog->Checked = false;
}

void TfrmMain::CallbackDataFn(short* buf, int size)
{
	ScopedLock<Mutex> lock(mutexDeviceData);
	data.clear();
	data.reserve(size/2);
	for (int i=0; i<size/2; i++)
	{
		data.push_back(buf[i]);
	}
	bDataUpdate = true;
}

void __fastcall TfrmMain::ActionAcquireStartExecute(TObject *Sender)
{
	if (device)
		device->AcquireStart();
	else
		MessageBox(this->Handle, "No device dll selected.\n"
			"Put dll's in \\device subdir and select with Settings/Common Settings.",
			this->Caption.c_str(), MB_ICONEXCLAMATION);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionAcquireStopExecute(TObject *Sender)
{
	if (device)
		device->AcquireStop();
}
//---------------------------------------------------------------------------

