/** \file MiniscopeDeviceInterface.cpp
*/

/* Copyright (C) 2008-2009 Tomasz Ostrowski <tomasz.o.ostrowski at gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.

 * Miniscope is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//---------------------------------------------------------------------------

#ifndef MiniscopeDeviceInterfaceH
#define MiniscopeDeviceInterfaceH
//---------------------------------------------------------------------------

#include "MiniscopeDevice.h"
#include "MiniscopeDeviceCapabilities.h"
#include <System.hpp>
#include <Forms.hpp>
#include <vector>
#include <set>

/** \brief Info about dll device library */
struct DllInfo
{
	AnsiString name;
	struct S_DEVICE_DLL_INTERFACE ver;
};

class DeviceInterface
{
private:

	/** \brief Verify if dll looks correctly and fill dllinfo for valid dll
		\return true if dll is valid
	*/
	static bool VerifyDll(AnsiString filename, struct DllInfo* const dllinfo);
	/** \brief Add dll to list
	*/
	static void AddDll(const struct DllInfo &dllinfo);

    static std::set<class DeviceInterface*> instances;

	AnsiString filename;	///< dll name (with full path)
	HINSTANCE hInstance;	///< dll instance ptr
	// dll function pointers
	pfGetMiniscopeInterfaceDescription dllGetMiniscopeInterfaceDescription;
	pfSetCallbacks dllSetCallbacks;
	pfShowSettings dllShowSettings;
	pfGetDeviceCapabilities dllGetDeviceCapabilities;
	pfSetSensitivity dllSetSensitivity;
	pfSetSamplingPeriod dllSetSamplingPeriod;
	pfSetTriggerSource dllSetTriggerSource;
	pfSetCouplingType dllSetCouplingType;
	pfSetBufferSize dllSetBufferSize;
	pfSetOffset dllSetOffset;
	pfConnect dllConnect;
	pfDisconnect dllDisconnect;
	pfAcquireStart dllAcquireStart;
	pfAcquireStop dllAcquireStop;
	pfManualTrigger dllManualTrigger;

	typedef void (__closure *CallbackData)(short*, int);


public:
	/** \brief Make list of valid dlls in supplied location
	*/
	static void EnumerateDlls(AnsiString dir);
	/** \brief Constructor
		\param asDllName	Name of dll used by object to communicate with device
	*/
	DeviceInterface(AnsiString asDllName);
	~DeviceInterface();
	int Load(void);			///< load dll (dll name was supplied to constructor)

	static std::vector<DllInfo> dlls;
	struct S_SCOPE_CAPABILITIES *capabilities;
    CallbackData callbackData;

	// dll callbacks
	static void __stdcall OnLog(void *cookie, char *szText);                   ///< called to generate log in parent application
	static void __stdcall OnConnect(void *cookie, int state, char *szMsgText); ///< called on connection/disconnection of device
	static void __stdcall OnData(void *cookie, void *buffer, unsigned int buffersize); ///< called to pass data (samples) to parent application
	static void __stdcall OnTrigger(void *cookie);

	void ShowSettings(void) {
		if (dllShowSettings)
			dllShowSettings(Application->Handle);
	}
	void AcquireStart(void) {
		if (dllAcquireStart)
			dllAcquireStart();
	}
	void AcquireStop(void) {
		if (dllAcquireStop)
			dllAcquireStop();
	}
	int Connect(void) {
		if (dllConnect)
			return dllConnect();
		return 0;
	}
	int Disconnect(void) {
		if (dllDisconnect)
			return dllDisconnect();
		return 0;
	}
};

#endif
