/** \file Display.h
    \brief Display drawing, Class representing View in MVC model
*/

/* Copyright (C) 2008-2009 Tomasz Ostrowski <tomasz.o.ostrowski at gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.

 * Miniscope is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef DisplayH
#define DisplayH

#include <vector>

enum E_DISPLAY_MODE
{
	MODE_OSCILLOSCOPE = 0,
    MODE_SPECTRUM
};

enum E_ELEMENT
{
	ELEMENT_BACKGROUND = 0,
	ELEMENT_VGRID,
	ELEMENT_HGRID,
	ELEMENT_PLOT,
	ELEMENT_CURSOR1,
	ELEMENT_CURSOR2
};

enum E_UPDATE_LEVEL
{
	UPDATE_ALL = 0,
	UPDATE_DATA,
	UPDATE_CURSOR,
	UPDATE_REPAINT
};

class Display
{
public:
	virtual ~Display() {};
	std::vector<float> *data;
	virtual void SetDiv(float fX, float fY);
	virtual void SetMode(const enum E_DISPLAY_MODE m) = 0;
	virtual const enum E_DISPLAY_MODE GetMode(void) {
		return mode;
	}
	virtual void Update(enum E_UPDATE_LEVEL level) = 0;
	/** \brief Resize viewport */
	virtual void ChangeSize(int w, int h) = 0;
	virtual void SetRange(int range) {
		iRange = range;
	}
	virtual void SetCursor1(int x, int y) = 0;
	virtual void SetCursor2(int x, int y) = 0;
	/** \brief Set viewport offset and zoom
		\param iCenterX	requested center of viewport (i.e. zoom click coordinate), -1 to omit
		\param iCenterY requested center of viewport (i.e. zoom click coordinate), -1 to omit
		\param fZoomX	X-axis scale
		\param fZoomY	Y-axis scale
	*/
	virtual void SetZoom(int iCenterX, int iCenterY, float fZoomX, float fZoomY) = 0;
	virtual void Move(int iStartX, int iStartY, int iStopX, int iStopY) = 0;
	/** \brief Transform screen coordinates to physical values */
	virtual void PosToPhysical(int x, int y, float &fValX, float &fValY) = 0;
protected:
	int iRange;					///< vertical range (range is from -iRange to iRange LSB)
	float fDivX, fDivY;			///< time and voltage per LSB
	enum E_DISPLAY_MODE mode;
	struct S_STATE
	{
		bool bGridX, bGridY;
		bool bGridLabels;			///< show value labels next to grid?
		float fZoomX, fZoomY;
		float fOffsetX, fOffsetY;	///< determines position of plot on canvas
		float fCursor1X, fCursor1Y, fCursor2X, fCursor2Y;
		int iCursor1X, iCursor1Y, iCursor2X, iCursor2Y;     //pixel values
	}   scope, spectrum;
	inline struct S_STATE& State(void) {
		if (mode == MODE_OSCILLOSCOPE)
			return scope;
		else
			return spectrum;
	}
};

#endif // DisplayH
